from pathlib import Path
from typing import Union

import cv2
import numpy as np
import torch
from torchvision import transforms as T


# TODO: draw utils..
def draw_results(
    image: Union[np.ndarray, str],
    results: list[dict],
    task: str,
    names: list[str],
):

    if task == "classification":

        def draw(image, results):
            image = cv2.putText(
                image,
                f"{names[results[0].get('class_id')]}",
                (10, 30),
                2,
                1.0,
                (0, 0, 255),
            )
            return image

    elif task == "object_detection":

        def draw(image, results):
            for result in results:
                x1, y1, w, h = result["bbox"]
                x2 = x1 + w
                y2 = y1 + h
                image = cv2.putText(
                    image,
                    f"{names[result.get('class_id')]}",
                    (int(x1), int(y1) - 2),
                    2,
                    1.0,
                    (0, 0, 255),
                )
                image = cv2.rectangle(
                    image,
                    (int(x1), int(y1)),
                    (int(x2), int(y2)),
                    (0, 0, 255),
                    2,
                )
            return image

    else:
        raise NotImplementedError(f"{task} draw function not implemented")

    if isinstance(image, str):
        image = cv2.imread(image)

    return draw(image, results)


def get_images(d, recursive: bool = True) -> list[str]:
    exp = "**/*" if recursive else "*"
    return list(
        map(
            str,
            list(Path(d).glob(exp + ".png"))
            + list(Path(d).glob(exp + ".jpg"))
            + list(Path(d).glob(exp + ".PNG"))
            + list(Path(d).glob(exp + ".JPG")),
        )
    )


# TODO: define image_info schema
def resize_image(
    image: np.ndarray, image_size: list[int], letter_box: bool = False
) -> list[np.ndarray, list[int], list[int], list[int], list[int]]:
    """Resize Image.

    Args:
        image (np.ndarray): opencv image.
        image_size (list[int]): image [width, height].
        letter_box (bool): letter box.

    Returns:
        list: resized_image, (ori_width, ori_height), (new_width, new_height), (left_pad, top_pad)
    """

    h, w = image.shape[:2]
    W, H = image_size

    if letter_box:
        if w > h:
            ratio = W / w
            w_ = W
            h_ = round(h * ratio)

            total_pad = w_ - h_
            top = total_pad // 2
            bottom = total_pad - top
            left, right = 0, 0

        else:
            ratio = H / h
            w_ = round(w * ratio)
            h_ = H

            total_pad = h_ - w_
            left = total_pad // 2
            right = total_pad - left
            top, bottom = 0, 0

        image = cv2.resize(image, (w_, h_), interpolation=cv2.INTER_CUBIC)
        image = cv2.copyMakeBorder(
            image, top, bottom, left, right, None, value=(114, 114, 114)
        )

        return image, (w, h), (w_, h_), (W, H), (left, top)

    else:
        w_, h_ = W, H
        image = cv2.resize(image, (w_, h_), interpolation=cv2.INTER_CUBIC)

        return image, (w, h), (w_, h_), (W, H), (0, 0)


def collate_fn(batch):
    images, infos = list(zip(*batch))
    return torch.stack(images, dim=0), infos


class ImageDataset:
    def __init__(
        self,
        image_dir: str,
        image_size: Union[int, list[int]],
        letter_box: bool = False,
    ):
        self.image_dir = image_dir
        if Path(self.image_dir).is_file():
            self.image_paths = [self.image_dir]
        else:
            self.image_paths = get_images(self.image_dir)

        self.image_size = (
            image_size
            if isinstance(image_size, list)
            else [image_size, image_size]
        )
        self.letter_box = letter_box

        self.transform = T.Compose(
            [
                T.ToTensor(),
            ]
        )

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, idx):
        image_path = self.image_paths[idx]

        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        (
            image,
            (ori_w, ori_h),
            (new_w, new_h),
            (input_w, input_h),
            (left_pad, top_pad),
        ) = resize_image(image, self.image_size, self.letter_box)

        return self.transform(image), {
            "image_path": image_path,
            "ori_shape": (ori_w, ori_h),
            "new_shape": (new_w, new_h),
            "input_shape": (input_w, input_h),
            "pad": (left_pad, top_pad),
        }

    def get_dataloader(self, batch_size: int, num_workers: int):
        return torch.utils.data.DataLoader(
            self,
            batch_size,
            num_workers=num_workers,
            collate_fn=collate_fn,
            shuffle=False,
            drop_last=False,
        )



"""
Ultralytics Hub
"""

from waffle_hub import get_installed_backend_version

BACKEND_NAME = "ultralytics"
BACKEND_VERSION = get_installed_backend_version(BACKEND_NAME)

from dataclasses import asdict
from pathlib import Path
from typing import Union

import torch
from torchvision import transforms as T
from ultralytics import YOLO
from waffle_utils.file import io

from waffle_hub.utils.image import ImageDataset

from .base_hub import BaseHub, InferenceContext, TrainContext
from .model.wrapper import ModelWrapper, ResultParser, get_parser


def get_preprocess(task: str, *args, **kwargs):

    if task == "classification":
        normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

        def preprocess(x):
            return normalize(x)

    elif task == "object_detection":
        normalize = T.Normalize([0, 0, 0], [1, 1, 1], inplace=True)

        def preprocess(x):
            return normalize(x)

    return preprocess


def get_postprocess(task: str, *args, **kwargs):

    if task == "classification":

        def inner(x: torch.Tensor):
            return [x]

    elif task == "object_detection":
        image_size = kwargs.get("image_size")
        image_size = (
            image_size
            if isinstance(image_size, list)
            else [image_size, image_size]
        )

        def inner(x: torch.Tensor):
            x = x[0]  # x[0]: prediction, x[1]: TODO: what is this...?
            x = x.transpose(1, 2)

            cxcywh = x[:, :, :4]
            cx, cy, w, h = torch.unbind(cxcywh, dim=-1)
            x1 = cx - w / 2
            y1 = cy - h / 2
            x2 = cx + w / 2
            y2 = cy + h / 2
            xyxy = torch.stack([x1, y1, x2, y2], dim=-1)

            xyxy[:, :, ::2] /= image_size[0]
            xyxy[:, :, 1::2] /= image_size[1]
            probs = x[:, :, 4:]
            confidences, class_ids = torch.max(probs, dim=-1)

            return xyxy, confidences, class_ids

    return inner


class UltralyticsHub(BaseHub):

    # Common
    MODEL_TYPES = ["yolov8"]
    MODEL_SIZES = list("nsmlx")  # TODO: generalize

    # Backend Specifics
    TASK_MAP = {
        "object_detection": "detect",
        "classification": "classify",
        # "segmentation": "segment"
        # "keypoint_detection": "pose"
    }
    TASK_SUFFIX = {
        "detect": "",
        "classify": "-cls",
        # "segment": "-seg",
    }

    def __init__(
        self,
        name: str,
        backend: str = None,
        version: str = None,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        classes: Union[list[dict], list] = None,
        root_dir: str = None,
    ):
        """Create Ultralytics Hub.

        Args:
            name (str): Hub name
            backend (str, optional): Backend name. See waffle_hub.get_backends(). Defaults to None.
            version (str, optional): Version. See waffle_hub.get_installed_backend_version(backend). Defaults to None.
            task (str, optional): Task Name. See UltralyticsHub.TASKS. Defaults to None.
            model_type (str, optional): Model Type. See UltralyticsHub.MODEL_TYPES. Defaults to None.
            model_size (str, optional): Model Size. See UltralyticsHub.MODEL_SIZES. Defaults to None.
            classes (Union[list[dict], list]): class dictionary or list. [{"supercategory": "name"}, ] or ["name",].
            root_dir (str, optional): Root directory of hub repository. Defaults to None.
        """
        super().__init__(
            name=name,
            backend=backend if backend else BACKEND_NAME,
            version=version if version else BACKEND_VERSION,
            task=task,
            model_type=model_type,
            model_size=model_size,
            classes=classes,
            root_dir=root_dir,
        )

    # Train Hook
    def on_train_start(self, ctx: TrainContext):
        # set data
        ctx.dataset_path: Path = Path(ctx.dataset_path)
        if self.backend_task_name in ["detect", "segment"]:
            if ctx.dataset_path.suffix not in [".yml", ".yaml"]:
                yaml_files = list(ctx.dataset_path.glob("*.yaml")) + list(
                    ctx.dataset_path.glob("*.yml")
                )
                if len(yaml_files) != 1:
                    raise FileNotFoundError(
                        f"Ambiguous data file. Detected files: {yaml_files}"
                    )
                ctx.dataset_path = Path(yaml_files[0]).absolute()
            else:
                ctx.dataset_path = ctx.dataset_path.absolute()
        elif self.backend_task_name == "classify":
            if not ctx.dataset_path.is_dir():
                raise ValueError(
                    f"Classification dataset should be directory. Not {ctx.dataset_path}"
                )
            ctx.dataset_path = ctx.dataset_path.absolute()
        ctx.dataset_path = str(ctx.dataset_path)

        # pretrained model
        ctx.pretrained_model = (
            ctx.pretrained_model
            if ctx.pretrained_model
            else self.model_type
            + self.model_size
            + self.TASK_SUFFIX[self.backend_task_name]
            + ".pt"
        )

    def training(self, ctx: TrainContext):
        try:
            model = YOLO(ctx.pretrained_model, task=self.backend_task_name)
            model.train(
                data=ctx.dataset_path,
                epochs=ctx.epochs,
                batch=ctx.batch_size,
                imgsz=ctx.image_size,
                rect=ctx.letter_box,
                device=ctx.device,
                workers=ctx.workers,
                seed=ctx.seed,
                verbose=ctx.verbose,
                project=self.hub_dir,
                name=self.RAW_TRAIN_DIR,
            )
            return model

        except Exception as e:
            if self.artifact_dir.exists():
                io.remove_directory(self.artifact_dir)
            raise e

    def on_train_end(self, ctx: TrainContext):
        io.copy_file(
            self.artifact_dir / "weights" / "best.pt",
            self.best_ckpt_file,
            create_directory=True,
        )
        io.copy_file(
            self.artifact_dir / "weights" / "last.pt",
            self.last_ckpt_file,
            create_directory=True,
        )
        io.copy_file(
            self.artifact_dir / "results.csv",
            self.metric_file,
            create_directory=True,
        )

    # Inference Hook
    def get_model(
        self, image_size: Union[int, list] = None, parser: ResultParser = None
    ):
        self.check_train_sanity()

        if image_size is None:
            train_config = io.load_yaml(self.train_config_file)
            image_size = train_config.get("image_size")

        # get adapt functions
        preprocess = get_preprocess(self.task)
        postprocess = get_postprocess(self.task, image_size=image_size)

        # get model
        model = ModelWrapper(
            model=YOLO(self.best_ckpt_file).model.eval(),
            preprocess=preprocess,
            postprocess=postprocess,
            parser=parser if parser else None,
        )

        return model

    def on_inference_start(self, ctx: InferenceContext):
        ctx.model = self.get_model(
            ctx.image_size, get_parser(self.task)(**asdict(ctx))
        )
        ctx.dataloader = ImageDataset(
            ctx.source, ctx.image_size, letter_box=ctx.letter_box
        ).get_dataloader(ctx.batch_size, ctx.workers)

    def on_inference_end(self, ctx: InferenceContext):
        pass

    # Evaluate Hook
    def evaluating(self):
        raise NotImplementedError



from pathlib import Path

import pytest
from waffle_utils.dataset import Dataset
from waffle_utils.file import io, network

from waffle_hub.hub import UltralyticsHub


@pytest.fixture
def dummy_dataset(tmpdir: Path):
    url = "https://github.com/snuailab/waffle_utils/raw/main/mnist.zip"

    dummy_zip_file = tmpdir / "mnist.zip"
    dummy_extract_dir = tmpdir / "extract"
    dummy_coco_root_dir = tmpdir / "extract/raw"
    dummy_coco_file = tmpdir / "extract/exports/coco.json"

    network.get_file_from_url(url, dummy_zip_file, create_directory=True)
    io.unzip(dummy_zip_file, dummy_extract_dir, create_directory=True)

    ds = Dataset.from_coco(
        "mnist", dummy_coco_file, Path(dummy_coco_root_dir), root_dir=tmpdir
    )
    ds.split_train_val(0.8)
    return ds


def test_ultralytics_detect_train_inference(
    tmpdir: Path, dummy_dataset: Dataset
):

    export_dir = dummy_dataset.export("yolo_detection")

    name = "test_det"

    hub = UltralyticsHub(
        name=name,
        task="object_detection",
        model_type="yolov8",
        model_size="n",
        classes=["1", "2"],
        root_dir=tmpdir,
    )
    hub = UltralyticsHub.load(name=name, root_dir=tmpdir)
    hub: UltralyticsHub = UltralyticsHub.from_model_config(
        name=name,
        model_config_file=tmpdir / name / UltralyticsHub.MODEL_CONFIG_FILE,
        root_dir=tmpdir,
    )
    hub.train(
        dataset_path=export_dir,
        epochs=1,
        batch_size=4,
        image_size=32,
        pretrained_model=None,
        device="cpu",
    )
    assert hub.check_train_sanity()

    inference_dir = hub.inference(
        source=export_dir,
        device="cpu",
    )


def test_ultralytics_classify_train(tmpdir: Path, dummy_dataset: Dataset):

    name = "test_cls"

    export_dir = dummy_dataset.export("yolo_classification")
    hub = UltralyticsHub(
        name=name,
        backend="ultralytics",
        task="classification",
        model_type="yolov8",
        model_size="n",
        classes=["1", "2"],
        root_dir=tmpdir,
    )
    hub.train(
        dataset_path=export_dir,
        epochs=1,
        batch_size=4,
        image_size=28,
        pretrained_model=None,
        device="cpu",
    )
    assert hub.check_train_sanity()

    inference_dir = hub.inference(
        source=export_dir,
        device="cpu",
    )



"""_summary_

Raises:
    NotImplementedError: _description_
    ValueError: _description_

Returns:
    _type_: _description_
"""

import torch
from torchvision.ops import batched_nms


class PreprocessFunction:
    pass


class PostprocessFunction:
    pass


class ResultParser:
    pass


class ClassificationResultParser(ResultParser):
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, results: list[torch.Tensor], *args, **kwargs):
        parseds = []

        results = results[0]  # TODO: multi label
        scores, class_ids = results.cpu().topk(results.shape[1], dim=-1)
        results = torch.cat(
            [class_ids.unsqueeze(-1), scores.unsqueeze(-1)], dim=-1
        )
        for result in results:
            parsed = []
            for class_id, score in result:
                parsed.append(
                    {"class_id": int(class_id), "score": float(score)}
                )
            parseds.append(parsed)
        return parseds


class ObjectDetectionResultParser(ResultParser):
    def __init__(
        self,
        confidence_threshold: float = 0.25,
        iou_threshold: float = 0.5,
        *args,
        **kwargs
    ):
        self.confidence_threshold = confidence_threshold
        self.iou_threshold = iou_threshold

    def __call__(
        self,
        results: list[torch.Tensor],
        image_infos: list[dict],
        *args,
        **kwargs
    ):
        parseds = []

        bboxes_batch, confs_batch, class_ids_batch = results
        for bboxes, confs, class_ids, image_info in zip(
            bboxes_batch, confs_batch, class_ids_batch, image_infos
        ):

            mask = confs > self.confidence_threshold

            bboxes, confs, class_ids = (
                bboxes[mask],
                confs[mask],
                class_ids[mask],
            )
            idxs = batched_nms(bboxes, confs, class_ids, self.iou_threshold)

            bboxes = bboxes[idxs, :].cpu()
            confs = confs[idxs].cpu()
            class_ids = class_ids[idxs].cpu()

            W, H = image_info.get("input_shape")
            left_pad, top_pad = image_info.get("pad")
            ori_w, ori_h = image_info.get("ori_shape")
            new_w, new_h = image_info.get("new_shape")

            parsed = []
            for (x1, y1, x2, y2), conf, class_id in zip(
                bboxes, confs, class_ids
            ):

                x1 = max(float((x1 * W - left_pad) / new_w * ori_w), 0)
                y1 = max(float((y1 * H - top_pad) / new_h * ori_h), 0)
                x2 = min(float((x2 * W - left_pad) / new_w * ori_w), ori_w)
                y2 = min(float((y2 * H - top_pad) / new_h * ori_h), ori_h)

                parsed.append(
                    {
                        "bbox": [x1, y1, x2 - x1, y2 - y1],
                        "score": float(conf),
                        "class_id": int(class_id),
                    }
                )
            parseds.append(parsed)
        return parseds


def get_parser(task: str):
    if task == "classification":
        return ClassificationResultParser
    elif task == "object_detection":
        return ObjectDetectionResultParser


class ModelWrapper(torch.nn.Module):
    def __init__(
        self,
        model: torch.nn.Module,
        preprocess: PreprocessFunction,
        postprocess: PostprocessFunction,
        parser: ResultParser = None,
    ):
        """
        Model Wrapper.
        Use this wrapper when inference, export.

        Args:
            model (torch.nn.Module): model
            preprocess (PreprocessFunction):
                Preprocess Function that
                recieves [batch, channel, height, width],
                and
                outputs [batch, channel, height, width].

            postprocess (PostprocessFunction):
                Postprocess Function that
                recieves model raw output,
                and
                outputs results that fit with our convention.

                Classification:
                    [
                        [batch, class_num],
                    ]  # scores per attribute
                Detection:
                    [
                        [batch, bbox_num, 4(x1, y1, x2, y2)],  # bounding box
                        [batch, bbox_num],  # confidence
                        [batch, bbox_num],  # class id
                    ]
                Segmentation:
                    # TODO: segmentation support
        """
        super().__init__()
        self.model = model
        self.preprocess = preprocess
        self.postprocess = postprocess
        self.parser = parser

    def forward(self, x, image_infos: list[dict] = None, *args, **kwargs):
        x = self.preprocess(x)
        x = self.model(x)
        x = self.postprocess(x)
        if self.parser is not None:
            x = self.parser(x, image_infos=image_infos, *args, **kwargs)
        return x



from dataclasses import dataclass


@dataclass
class Model:
    name: str
    backend: str
    version: str
    task: str
    model_type: str
    model_size: str
    classes: list


@dataclass
class Train:
    image_size: int
    letter_box: bool
    batch_size: int
    pretrained_model: str
    seed: int


# TODO: Should be moved to waffle_utils. Define Prediction schema for now.
@dataclass
class Prediction:
    image_path: str
    predictions: list[dict]


@dataclass
class DetectionPrediction:
    bbox: list[float]
    class_name: str
    confidence: float
    segment: list[float] = list


@dataclass
class ClassificationPrediction:
    score: float
    class_name: str



# TODO: add cli tool
# import typer
# from rich import print

# from waffle_hub.hub import UltralyticsHub

# from waffle_utils.dataset import Dataset

# app = typer.Typer()

# BACKEND_MAP = {
#     "ultralytics": UltralyticsHub,
# }

# @app.command(name="create", )
# def _create(
#     backend: str = typer.Option(..., help='One of [' + ', '.join(BACKEND_MAP.keys()) + ']'),
#     name: str = type.Option(..., help="model name"),
#     task: str = type.Option(..., help="task"),
#     model_type: str = type.Option(..., help="model type"),
#     model_size: str = type.Option(..., help="model size"),
#     pretrained_model: str = type.Option(None, help="pretrained model"),
#     model_root_dir: str = type.Option(None, help="pretrained model"),
#     dataset_
#     epochs: int,
#     batch_size: int,
#     image_size: int,
#     device: str = "0",
#     workers: int = 2,
#     seed: int = 0,
#     verbose: bool = True,
# ):
#     if backend not in BACKEND_MAP:
#         raise ValueError(f"""
#         Backend {backend} is not supported.
#         Choose one of {list(BACKEND_MAP.keys())}
#         """)

#     # hub = BACKEND_MAP[backend](
#     #     name=name,
#     #     task=task
#     # )


# if __name__ == "__main__":
#     app()



__version__ = "0.1.0"

import importlib
import warnings
from collections import OrderedDict

from tabulate import tabulate

# pytorch install check (necessary installation)
pytorch_versions = ["1.13.1"]
try:
    import torch

    if torch.__version__ not in pytorch_versions:
        warnings.warn(
            f"""
            torch {torch.__version__} has not been tested.
            We recommend you to use one of {pytorch_versions}
            """
        )
except ModuleNotFoundError as e:
    # TODO: Generalize install strings
    strings = [
        "  - conda install pytorch==1.12.1 torchvision==0.13.1 torchaudio==0.12.1 cudatoolkit=11.3 -c pytorch",
        "  - pip install torch==1.12.1+cu113 torchvision==0.13.1+cu113 torchaudio==0.12.1 --extra-index-url https://download.pytorch.org/whl/cu113",
        "  - pip install torch==1.12.1 torchvision==0.13.1 torchaudio==0.12.1",
    ]

    e.msg = "Need to install torch\n" + "\n".join(strings)
    raise e

# backend supports
_backends = OrderedDict(
    {
        "ultralytics": ["8.0.54"],
    }
)


def get_backends() -> dict:
    return _backends


def get_installed_backend_version(backend: str) -> str:

    backends = get_backends()
    versions = backends[backend]

    if backend not in backends:
        raise ModuleNotFoundError(
            f"{backend} is not supported.\n Available backends {list(backends.keys())}"
        )

    try:
        module = importlib.import_module(backend)
        if module.__version__ not in versions:
            warnings.warn(
                f"""
                {backend} {module.__version__} has not been tested.
                We recommend you to use one of {versions}
                """
            )
        return module.__version__

    except ModuleNotFoundError as e:

        install_queries = "\n".join(
            [f"- pip install {backend}=={version}" for version in versions]
        )

        e.msg = f"""
            Need to install {backend}.
            Tested versions:
            {install_queries}
            """
        raise e


def get_available_backends() -> str:
    """Available backends"""
    backends = get_backends()

    table_data = []
    for name, versions in backends.items():
        for i, version in enumerate(versions):
            table_data.append([name if i == 0 else "", version])

    table = tabulate(
        table_data,
        headers=["Backend", "Version"],
        tablefmt="simple_outline",
    )

    return table



"""
Base Hub Class
Do not use this Class directly.
Use {Backend}Hub instead.
"""

import logging
from abc import abstractmethod
from dataclasses import asdict, dataclass
from functools import cached_property
from pathlib import Path
from typing import Union

import cv2
import torch
import tqdm
from waffle_utils.file import io
from waffle_utils.utils import type_validator

from waffle_hub.schemas.configs import Model, Train
from waffle_hub.utils.image import draw_results

logger = logging.getLogger(__name__)


class ConfigContext:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        return


@dataclass
class TrainContext(ConfigContext):
    dataset_path: str
    epochs: int
    batch_size: int
    image_size: int
    letter_box: bool
    pretrained_model: str
    device: str
    workers: int
    seed: int
    verbose: bool


@dataclass
class InferenceContext(ConfigContext):
    source: str
    batch_size: int
    recursive: bool
    image_size: int
    letter_box: bool
    confidence_threshold: float
    iou_thresold: float
    half: bool
    workers: int
    device: str
    draw: bool

    model = None
    dataloader = None


@dataclass
class ExportContext(ConfigContext):
    image_size: Union[int, list]
    batch_size: int
    input_name: list[str]
    output_name: list[str]
    opset_version: int


class BaseHub:

    TASKS = [
        "object_detection",
        "classification",
        "segmentation",
        "keypoint_detection",
    ]
    MODEL_TYPES = []
    MODEL_SIZES = []

    # directory settings
    DEFAULT_ROOT_DIR = Path("./hubs")

    RAW_TRAIN_DIR = Path("artifacts")

    INFERENCE_DIR = Path("inferences")
    EVALUATION_DIR = Path("evaluations")
    EXPORT_DIR = Path("exports")

    # config files
    CONFIG_DIR = Path("configs")
    MODEL_CONFIG_FILE = CONFIG_DIR / "model.yaml"
    TRAIN_CONFIG_FILE = CONFIG_DIR / "train.yaml"

    # train results
    LAST_CKPT_FILE = "weights/last_ckpt.pt"
    BEST_CKPT_FILE = "weights/best_ckpt.pt"  # TODO: best metric?
    METRIC_FILE = "metrics.csv"

    # export results
    ONNX_FILE = "weights/model.onnx"

    def __init__(
        self,
        name: str,
        backend: str = None,
        version: str = None,
        task: str = None,
        model_type: str = None,
        model_size: str = None,
        classes: Union[list[dict], list] = None,
        root_dir: str = None,
    ):
        self.name: str = name
        self.task: str = task
        self.model_type: str = model_type
        self.model_size: str = model_size
        self.classes: list[dict] = classes
        self.root_dir: Path = Path(root_dir) if root_dir else None

        self.backend: str = backend
        self.version: str = version

        # check task supports
        self.backend_task_name = self.TASK_MAP.get(self.task, None)
        if self.backend_task_name is None:
            io.remove_directory()
            raise ValueError(
                f"{self.task} is not supported with {self.backend}"
            )

        try:
            # save model config
            model_config = Model(
                name=self.name,
                backend=self.backend,
                version=self.version,
                task=self.task,
                model_type=self.model_type,
                model_size=self.model_size,
                classes=self.classes,
            )
            io.save_yaml(
                asdict(model_config),
                self.model_config_file,
                create_directory=True,
            )
        except Exception as e:
            raise e

    @classmethod
    def load(cls, name: str, root_dir: str = None) -> "BaseHub":
        """Load Hub by name.

        Args:
            name (str): hub name.
            root_dir (str, optional): hub root directory. Defaults to None.

        Raises:
            FileNotFoundError: if hub is not exist in root_dir

        Returns:
            Hub: Hub instance
        """
        model_config_file = (
            Path(root_dir if root_dir else BaseHub.DEFAULT_ROOT_DIR)
            / name
            / BaseHub.MODEL_CONFIG_FILE
        )
        if not model_config_file.exists():
            raise FileNotFoundError(
                f"Model[{name}] does not exists. {model_config_file}"
            )
        model_config = io.load_yaml(model_config_file)
        return cls(**model_config)

    @classmethod
    def from_model_config(
        cls, name: str, model_config_file: str, root_dir: str = None
    ) -> "BaseHub":
        """Create new Hub with model config.

        Args:
            name (str): hub name.
            model_config_file (str): model config yaml file.
            root_dir (str, optional): hub root directory. Defaults to None.

        Returns:
            Hub: New Hub instance
        """
        model_config = io.load_yaml(model_config_file)
        return cls(
            **{
                **model_config,
                "name": name,
                "root_dir": root_dir,
            }
        )

    # properties
    @property
    def name(self) -> str:
        """Hub name"""
        return self.__name

    @name.setter
    @type_validator(str)
    def name(self, v):
        self.__name = v

    @property
    def root_dir(self) -> Path:
        """Root Directory"""
        return self.__root_dir

    @root_dir.setter
    @type_validator(Path, strict=False)
    def root_dir(self, v):
        self.__root_dir = Path(v) if v else BaseHub.DEFAULT_ROOT_DIR

    @property
    def task(self) -> str:
        """Task Name"""
        return self.__task

    @task.setter
    @type_validator(str)
    def task(self, v):
        if v not in self.TASKS:
            raise ValueError(
                f"Task {v} is not supported. Choose one of {self.TASKS}"
            )
        self.__task = v

    @property
    def model_type(self) -> str:
        """Model Type"""
        return self.__model_type

    @model_type.setter
    @type_validator(str)
    def model_type(self, v):
        if v not in self.MODEL_TYPES:
            raise ValueError(
                f"Model Type {v} is not supported. Choose one of {self.MODEL_TYPES}"
            )
        self.__model_type = v

    @property
    def model_size(self) -> str:
        """Model Size"""
        return self.__model_size

    @model_size.setter
    @type_validator(str)
    def model_size(self, v):
        if v not in self.MODEL_SIZES:
            raise ValueError(
                f"Model Size {v} is not supported. Choose one of {self.MODEL_SIZES}"
            )
        self.__model_size = v

    @property
    def backend(self) -> str:
        """Backend name"""
        return self.__backend

    @backend.setter
    @type_validator(str)
    def backend(self, v):
        self.__backend = v

    @property
    def version(self) -> str:
        """Version"""
        return self.__version

    @version.setter
    @type_validator(str)
    def version(self, v):
        self.__version = v

    @property
    def classes(self) -> list[dict]:
        return self.__classes

    @classes.setter
    @type_validator(list)
    def classes(self, v):
        if isinstance(v[0], str):
            v = [{"supercategory": "object", "name": n} for n in v]
        self.__classes = v

    @cached_property
    def hub_dir(self) -> Path:
        """Hub(Model) Directory"""
        return self.root_dir / self.name

    @cached_property
    def artifact_dir(self) -> Path:
        """Artifact Directory. This is raw output of each backend."""
        return self.hub_dir / BaseHub.RAW_TRAIN_DIR

    @cached_property
    def inference_dir(self) -> Path:
        """Inference Results Directory"""
        return self.hub_dir / BaseHub.INFERENCE_DIR

    @cached_property
    def evaluation_dir(self) -> Path:
        """Evaluation Results Directory"""
        return self.hub_dir / BaseHub.EVALUATION_DIR

    @cached_property
    def export_dir(self) -> Path:
        """Export Results Directory"""
        return self.hub_dir / BaseHub.EXPORT_DIR

    @cached_property
    def model_config_file(self) -> Path:
        """Model Config yaml File"""
        return self.hub_dir / BaseHub.MODEL_CONFIG_FILE

    @cached_property
    def train_config_file(self) -> Path:
        """Train Config yaml File"""
        return self.hub_dir / BaseHub.TRAIN_CONFIG_FILE

    @cached_property
    def best_ckpt_file(self) -> Path:
        """Best Checkpoint File"""
        return self.hub_dir / BaseHub.BEST_CKPT_FILE

    @cached_property
    def onnx_file(self) -> Path:
        """Best Checkpoint File"""
        return self.hub_dir / BaseHub.ONNX_FILE

    @cached_property
    def last_ckpt_file(self) -> Path:
        """Last Checkpoint File"""
        return self.hub_dir / BaseHub.LAST_CKPT_FILE

    @cached_property
    def metric_file(self) -> Path:
        """Metric Csv File"""
        return self.hub_dir / BaseHub.METRIC_FILE

    # common functions
    def delete_artifact(self):
        """Delete Artifact Directory. It can be trained again."""
        io.remove_directory(self.artifact_dir)

    def check_train_sanity(self) -> bool:
        """Check if all essential files are exist.

        Returns:
            bool: True if all files are exist else False
        """
        if not (
            self.model_config_file.exists()
            and self.best_ckpt_file.exists()
            # and self.last_ckpt_file.exists()
        ):
            raise FileNotFoundError("Train first! hub.train(...).")
        return True

    # Train Hook
    def before_train(self, ctx: TrainContext):
        if self.artifact_dir.exists():
            raise FileExistsError(
                "Train artifacts already exist. Remove artifact to re-train (hub.delete_artifact())."
            )

    def on_train_start(self, ctx: TrainContext):
        pass

    def save_train_config(self, ctx: TrainContext):
        io.save_yaml(
            asdict(
                Train(
                    image_size=ctx.image_size,
                    letter_box=ctx.letter_box,
                    batch_size=ctx.batch_size,
                    pretrained_model=ctx.pretrained_model,
                    seed=ctx.seed,
                )
            ),
            self.train_config_file,
            create_directory=True,
        )

    def training(self, ctx: TrainContext):
        pass

    def on_train_end(self, ctx: TrainContext):
        pass

    def after_train(self, ctx: TrainContext):
        pass

    def train(
        self,
        dataset_path: str,
        epochs: int,
        batch_size: int,
        image_size: int,
        letter_box: bool = False,
        pretrained_model: str = None,
        device: str = "0",
        workers: int = 2,
        seed: int = 0,
        verbose: bool = True,
    ) -> str:
        """Start Train

        Args:
            dataset_path (str): Dataset Path. Recommend to use result of waffle_utils.dataset.Dataset.export.
            epochs (int): total epochs
            batch_size (int): batch size
            image_size (int): image size
            letter_box (bool): letter box preprocess. Defaults to False.
            pretrained_model (str, optional): pretrained model file. Defaults to None.
            device (str, optional): gpu device. Defaults to "0".
            workers (int, optional): num workers. Defaults to 2.
            seed (int, optional): random seed. Defaults to 0.
            verbose (bool, optional): verbose. Defaults to True.

        Raises:
            FileExistsError: if trained artifact exists.
            FileNotFoundError: if can not detect appropriate dataset.
            ValueError: if can not detect appropriate dataset.
            e: something gone wrong with ultralytics

        Returns:
            str: hub directory
        """

        with TrainContext(
            dataset_path=dataset_path,
            epochs=epochs,
            batch_size=batch_size,
            image_size=image_size,
            letter_box=letter_box,
            pretrained_model=pretrained_model,
            device="cpu" if device == "cpu" else f"cuda:{device}",
            workers=workers,
            seed=seed,
            verbose=verbose,
        ) as ctx:
            self.before_train(ctx)
            self.on_train_start(ctx)
            self.save_train_config(ctx)
            self.training(ctx)
            self.on_train_end(ctx)
            self.after_train(ctx)

        return str(self.hub_dir)

    # Inference Hook
    def get_model(self):
        raise NotImplementedError

    def before_inference(self, ctx: InferenceContext):
        self.check_train_sanity()

        # overwrite training config
        train_config = io.load_yaml(self.train_config_file)
        if ctx.image_size is None:
            ctx.image_size = train_config.get("image_size")
        if ctx.letter_box is None:
            ctx.letter_box = train_config.get("letter_box")

    def on_inference_start(self, ctx: InferenceContext):
        pass

    def inferencing(self, ctx: InferenceContext) -> str:
        model = ctx.model.to(ctx.device)
        dataloader = ctx.dataloader
        device = ctx.device

        for images, image_infos in tqdm.tqdm(dataloader):
            result_batch = model(images.to(device), image_infos)
            for results, image_info in zip(result_batch, image_infos):
                image_path = image_info.get("image_path")

                relpath = Path(image_path).relative_to(ctx.source)
                if relpath == Path("."):
                    relpath = Path(
                        "test.png"
                    )  # TODO: path correction for none directory source.
                io.save_json(
                    results,
                    self.inference_dir
                    / "results"
                    / relpath.with_suffix(".json"),
                    create_directory=True,
                )
                if ctx.draw:
                    draw = draw_results(
                        image_path,
                        results,
                        task=self.task,
                        names=[x["name"] for x in self.classes],
                    )
                    draw_path = (
                        self.inference_dir
                        / "draw"
                        / relpath.with_suffix(".png")
                    )
                    io.make_directory(draw_path.parent)
                    cv2.imwrite(str(draw_path), draw)

        return results  # TODO: return what?

    def on_inference_end(self, ctx: InferenceContext):
        pass

    def after_inference(self, ctx: InferenceContext):
        pass

    def inference(
        self,
        source: str,
        recursive: bool = True,
        image_size: int = None,
        letter_box: bool = None,
        batch_size: int = 4,
        confidence_threshold: float = 0.25,
        iou_thresold: float = 0.5,
        half: bool = False,
        workers: int = 2,
        device: str = "0",
        draw: bool = False,
    ) -> str:
        """Start Inference

        Args:
            source (str): dataset source. image file or image directory. TODO: video
            recursive (bool, optional): get images from directory recursively. Defaults to True.
            image_size (int, optional): inference image size. None for same with train_config (recommended).
            letter_box (bool, optional): letter box preprocess. None for same with train_config (recommended).
            batch_size (int, optional): batch size. Defaults to 4.
            conf_thres (float, optional): confidence threshold. Defaults to 0.25.
            iou_thres (float, optional): iou threshold. Defaults to 0.7.
            half (bool, optional): fp16 inference. Defaults to False.
            device (str, optional): gpu device. Defaults to "0".
            draw (bool, optional): save draw or not. Defaults to False.

        Raises:
            FileNotFoundError: if can not detect appropriate dataset.
            e: something gone wrong with ultralytics

        Returns:
            str: inference result directory
        """
        self.check_train_sanity()

        with InferenceContext(
            source=source,
            batch_size=batch_size,
            recursive=recursive,
            image_size=image_size,
            letter_box=letter_box,
            confidence_threshold=confidence_threshold,
            iou_thresold=iou_thresold,
            half=half,
            workers=workers,
            device="cpu" if device == "cpu" else f"cuda:{device}",
            draw=draw,
        ) as ctx:

            self.before_inference(ctx)
            self.on_inference_start(ctx)
            self.inferencing(ctx)
            self.on_inference_end(ctx)
            self.after_inference(ctx)

        return str(self.inference_dir)

    # Export Hook
    def export(
        self,
        image_size: int = None,
        batch_size: int = 1,
        opset_version: int = 11,
    ) -> str:
        """Export Model

        Args:
            image_size (int, optional): inference image size. None for same with train_config (recommended).
            batch_size (int, optional): dynamic batch size. Defaults to 16.
            opset_version (int, optional): onnx opset version. Defaults to 11.

        Returns:
            str: export onnx file path
        """
        self.check_train_sanity()

        train_config = Train(**io.load_yaml(self.train_config_file))

        image_size = image_size if image_size else train_config.image_size
        image_size = (
            [image_size, image_size]
            if isinstance(image_size, int)
            else image_size
        )

        model = self.get_model(train_config.image_size)

        input_name = ["inputs"]
        if self.task == "object_detection":
            output_names = ["bbox", "conf", "class_id"]
        elif self.task == "classification":
            output_names = ["predictions"]
        else:
            raise NotImplementedError(
                f"{self.task} does not support export yet."
            )

        dummy_input = torch.randn(batch_size, 3, *image_size)

        torch.onnx.export(
            model,
            dummy_input,
            str(self.onnx_file),
            input_names=input_name,
            output_names=output_names,
            opset_version=opset_version,
            dynamic_axes={
                name: {0: "batch_size"} for name in input_name + output_names
            },
        )

        return str(self.onnx_file)

    @abstractmethod
    def evaluate(self):
        raise NotImplementedError
